import mimetypes
import os
from ast import literal_eval

import requests


class Api(object):

    def __init__(self, url, private_token, proxy=None):
        self.url = url
        mimetypes.add_type('text/plain', '.log')
        mimetypes.add_type('application/octet-stream', '')
        self.request = requests.Session()
        self.request.headers = {'Authorization': 'PrivateToken ' + private_token}
        self.request.proxies = literal_eval(proxy) if proxy else {}

    # AutoTests
    def create_autotest(self, json):
        response = self.request.post(
            self.url + '/api/v2/autoTests',
            json=json
        )
        print(f"Autotest: {json['name']}")
        if response.status_code != 201:
            raise Exception(f"Create autotest error: {response.json()['error']['key']}")
        print('\nCreate autotest passed!')
        return response.json()['id']

    def link_autotest(self, autotest_id, workitem_id):
        response = self.request.post(
            f'{self.url}/api/v2/autoTests/{autotest_id}/workItems',
            json={'id': workitem_id}
        )
        if response.status_code == 204:
            print('Link autoTest with workItems passed!')
        else:
            print(f"Link autoTest with workItems error: {response.json()['error']['key']}")

    def get_autotest(self, external_id, project_id):
        response = self.request.get(
            f'{self.url}/api/v2/autoTests?projectId={project_id}&externalId={external_id}'
        )
        if response.status_code != 200:
            raise Exception(f"Get autoTest error: {response.json()['error']['key']}")
        print('\nGet autoTest passed!')
        return response

    def update_autotest(self, json):
        response = self.request.put(
            self.url + '/api/v2/autoTests',
            json=json
        )
        print(f"AutoTest: {json['name']}")
        if response.status_code == 204:
            print('Update passed!')
        else:
            raise Exception(f"Update error: {response.json()['error']['key']}")

    # TestRuns
    def create_testrun(self, json):
        response = self.request.post(
            self.url + '/api/v2/testRuns',
            json=json
        )
        if response.status_code != 201:
            raise Exception(f"Create testRun error: {response.json()['error']['key']}")
        print('Create testRun passed!')
        return response.json()['id']

    def get_testrun(self, testrun_id):
        response = self.request.get(
            f'{self.url}/api/v2/testRuns/{testrun_id}'
        )
        if response.status_code != 200:
            raise Exception(f"Get testRun error: {response.json()['error']['key']}")
        print('Get testRun passed!')
        return response.json()['projectId'], response.json()['testResults']

    def set_results_for_testrun(self, testrun_id, json):
        response = self.request.post(
            f'{self.url}/api/v2/testRuns/{testrun_id}/testResults',
            json=json
        )
        if response.status_code == 200:
            print('Set results passed!')
        else:
            raise Exception(f"Set results error: {response.json()['error']['key']}")

    def testrun_activity(self, testrun_id, action):
        response = self.request.post(
            f'{self.url}/api/v2/testRuns/{testrun_id}/{action}'
        )
        if response.status_code == 204:
            print(f'TestRun {action} passed!')
        else:
            raise Exception(f"TestRun {action} error: {response.json()['error']['key']}")

    def load_attachment(self, file):
        response = self.request.post(
            f'{self.url}/api/Attachments',
            files={
                'file': (os.path.basename(file.name),
                         file,
                         mimetypes.guess_type(file.name)[0])
            }
        )
        if response.status_code == 201:
            print(f'Attachment {file.name} loaded!')
            return response.json()['id']
        else:
            print(f"Attachment {file.name} error: {response.json()['error']['key']}")
            return None



from pluggy import PluginManager


class TestITPluginManager:
    plugin_manager = None

    @classmethod
    def get_plugin_manager(cls):
        if not cls.plugin_manager:
            cls.plugin_manager = PluginManager('testit')
        return cls.plugin_manager

    @classmethod
    def __getattr__(cls, attribute):
        return getattr(cls.get_plugin_manager(), attribute)



import configparser
import inspect
import os
import pickle
from datetime import datetime

import pytest
import testit_pytest
from testit_pytest.api import Api
from testit_pytest.json_fixture import JSONFixture
from testit_pytest.utils import (
    configurations_parser,
    search_in_environ,
    step,
    url_check,
    uuid_check
)


class TestITListener(object):
    # TODO: need to rework for normal read
    def __init__(self, testrun, url, private_token, proxy=None):
        if testrun:
            self.testrun_id = uuid_check(testrun)

            if not url:
                print('URL was not found!')
                raise SystemExit
            url_check(url)

            if not private_token:
                print('Private token was not found!')
                raise SystemExit
        else:
            path = os.path.abspath('')
            root = path[:path.index(os.sep)]
            while not os.path.isfile(
                    f'{path}{os.sep}connection_config.ini') and path != root:
                path = path[:path.rindex(os.sep)]
            path = f'{path}{os.sep}connection_config.ini'
            if not os.path.isfile(path):
                print('File connection_config.ini was not found!')
                raise SystemExit
            parser = configparser.RawConfigParser()
            parser.read(path)

            url = url_check(search_in_environ(parser.get('testit',
                                                         'url')))

            private_token = search_in_environ(parser.get('testit',
                                                         'privatetoken'))
            try:
                proxy = search_in_environ(parser.get('debug',
                                                     'testit_proxy'))
            except configparser.NoSectionError as nse:
                print(f"Section not set, cause: {nse}")
            except configparser.NoOptionError as noe:
                print(f"Option not set, cause: {noe}")

            if parser.has_option('testit', 'testrunID'):
                self.testrun_id = uuid_check(search_in_environ(
                    parser.get('testit', 'testrunID')))
            else:
                self.project_id = uuid_check(
                    search_in_environ(parser.get('testit', 'projectID')))
                self.configuration_id = uuid_check(
                    search_in_environ(parser.get('testit', 'configurationID')))

                if parser.has_option('testit', 'testrun_name'):
                    self.testrun_name = parser.get('testit', 'testrun_name')
        self.requests = Api(url, private_token, proxy=proxy)

    @pytest.hookimpl
    def pytest_configure(self, config):
        if not hasattr(config, "workerinput") and not hasattr(self, "testrun_id"):
            config.testrun_id = self.requests.create_testrun(
                JSONFixture.create_testrun(
                    self.project_id,
                    f'{self.testrun_name if hasattr(self, "testrun_name") else "LocalRun"} {datetime.today().strftime("%d %b %Y %H:%M:%S")}'))
            self.requests.testrun_activity(config.testrun_id, 'start')

    @pytest.hookimpl
    def pytest_configure_node(self, node):
        if not hasattr(self, "testrun_id"):
            node.workerinput["testrun_id"] = pickle.dumps(node.config.testrun_id)

    @pytest.hookimpl
    def pytest_collection_modifyitems(self, session, config, items):
        index = 0
        new_items = []
        plugin_info = config.pluginmanager.list_plugin_distinfo()

        for plugin, dist in plugin_info:
            if 'pytest_check' == dist.project_name:
                from pytest_check import check_methods
                self.pytest_check_get_failures = check_methods.get_failures
                break

        if not hasattr(self, "testrun_id"):
            self.testrun_id = pickle.loads(config.workerinput["testrun_id"]) if\
                hasattr(config, 'workerinput') else config.testrun_id

        self.project_id, data_autotests = self.requests.get_testrun(
        self.testrun_id)
        configurations_array = configurations_parser(data_autotests)

        for item in items:
            if hasattr(item.function, 'test_external_id'):
                if item.own_markers:
                    for mark in item.own_markers:
                        if mark.name == 'parametrize':
                            if not hasattr(item, 'array_parametrize_mark_id'):
                                item.array_parametrize_mark_id = []
                            item.array_parametrize_mark_id.append(
                                item.own_markers.index(mark))

                item.test_external_id = self.param_attribute_collector(
                    item.function.test_external_id,
                    item.own_markers,
                    item.array_parametrize_mark_id,
                    index) if hasattr(item,
                                      'array_parametrize_mark_id'
                                      ) else item.function.test_external_id

                item.index = index
                item_id = items.index(item)
                index = index + 1 if item_id + \
                                     1 < len(items) and item.originalname == \
                                     items[item_id + 1].originalname else 0

                if configurations_array:
                    if item.test_external_id in configurations_array:
                        item.test_configuration_id = configurations_array[
                            item.test_external_id]
                        new_items.append(item)
                else:
                    item_id = session.items.index(item)
                    session.items[item_id].test_configuration_id = self.configuration_id
        if configurations_array:
            if not new_items:
                self.requests.testrun_activity(self.testrun_id, 'stop')
                print('The specified tests were not found!')
                raise SystemExit
            session.items = new_items

    @pytest.hookimpl
    def pytest_runtest_protocol(self, item):
        if hasattr(item.function, 'test_external_id'):
            if not hasattr(item.function,
                           'test_displayname') and not item.function.__doc__:
                self.item = None
                raise Exception(
                    f'{item.originalname} must have @testit.displayName or documentation!')
            self.item = item
        else:
            self.item = None

    @pytest.hookimpl(hookwrapper=True)
    def pytest_fixture_setup(self, fixturedef):
        yield
        if hasattr(self, 'item') and self.item:
            steps_data, results_steps_data = step.get_steps_data()
            if fixturedef.scope == 'function':
                if not hasattr(self.item, 'test_method_setup_steps'):
                    self.item.test_method_setup_steps = []
                    self.item.test_method_setup_results_steps = []
                self.item.test_method_setup_steps += steps_data
                self.item.test_method_setup_results_steps += results_steps_data

    @pytest.hookimpl(hookwrapper=True, trylast=True)
    def pytest_runtest_call(self):
        yield
        if hasattr(self, 'item') and not self.item:
            return
        test_steps, test_results_steps = step.get_steps_data()
        self.item.test_steps = test_steps
        self.item.test_results_steps = test_results_steps

    @pytest.hookimpl
    def pytest_fixture_post_finalizer(self, fixturedef):
        if hasattr(self, 'item') and not self.item:
            return
        teardown_steps, teardown_results_steps = step.get_steps_data()
        if fixturedef.scope == 'function':
            if not hasattr(self.item, 'test_method_teardown_steps'):
                self.item.test_method_teardown_steps = []
                self.item.test_method_teardown_results_steps = []
            self.item.test_method_teardown_steps += teardown_steps
            self.item.test_method_teardown_results_steps += teardown_results_steps

    @pytest.hookimpl
    def pytest_runtest_logreport(self, report):
        if hasattr(self, 'item') and self.item:
            if report.failed or hasattr(report, 'wasxfail') \
                    and not report.passed or report.outcome == 'rerun':
                if not hasattr(self.item, 'test_traces'):
                    self.item.test_traces = ''
                self.item.test_traces += (report.longreprtext if report.longreprtext else
                    ''.join(['\n'.join(self.item.old_failures)]) +
                    f'\n{"-" * 60}\nFailed Checks: {len(self.item.old_failures)}') + f'\n{"-" * 60}\n'
                if hasattr(self.item, 'old_failures'):
                    del self.item.old_failures
            if not hasattr(self.item, 'test_duration'):
                self.item.test_duration = 0
            self.item.test_duration += report.duration * 1000

    @pytest.hookimpl
    def pytest_runtest_logfinish(self):
        if not hasattr(self, 'item') or not self.item:
            return
        data_item = self.attribute_collector(self.item)
        if data_item:
            autotest = self.requests.get_autotest(
                data_item['externalID'], self.project_id).json()
            if not autotest:
                autotest_id = self.requests.create_autotest(
                    JSONFixture.create_autotest(
                        data_item['externalID'],
                        self.project_id,
                        data_item['autoTestName'],
                        data_item['namespace'],
                        data_item['classname'],
                        data_item['links'],
                        data_item['steps'],
                        data_item['setUp'],
                        data_item['tearDown'],
                        data_item['title'],
                        data_item['description'],
                        data_item['labels']
                    )
                )
            else:
                autotest_id = autotest[0]['id']
                if data_item['testResult'] == 'Passed':
                    self.requests.update_autotest(
                        JSONFixture.update_autotest(
                            data_item['externalID'],
                            self.project_id,
                            data_item['autoTestName'],
                            data_item['namespace'],
                            data_item['classname'],
                            data_item['links'],
                            data_item['steps'],
                            data_item['setUp'],
                            data_item['tearDown'],
                            data_item['title'],
                            data_item['description'],
                            data_item['labels'],
                            autotest_id
                        )
                    )
                else:
                    self.requests.update_autotest(
                        JSONFixture.update_autotest(
                            data_item['externalID'],
                            self.project_id,
                            autotest[0]['name'],
                            autotest[0]['namespace'],
                            autotest[0]['classname'],
                            data_item['links'],
                            autotest[0]['steps'],
                            autotest[0]['setup'],
                            autotest[0]['teardown'],
                            autotest[0]['title'],
                            autotest[0]['description'],
                            autotest[0]['labels'],
                            autotest_id
                        )
                    )

            for workitem_id in data_item['workItemsID']:
                self.requests.link_autotest(autotest_id, workitem_id)

            self.requests.set_results_for_testrun(
                self.testrun_id,
                [JSONFixture.set_results_for_testrun(
                    data_item['externalID'],
                    data_item['configurationID'],
                    data_item['testResult'],
                    data_item['stepResults'],
                    data_item['setUpResults'],
                    data_item['tearDownResults'],
                    data_item['traces'],
                    data_item['resultLinks'],
                    data_item['duration'],
                    data_item['failureReasonName'],
                    data_item['message'],
                    data_item['parameters'],
                    data_item['attachments']
                )]
            )

    # TODO: move private methods into separate classes according to the purpose of the object
    @staticmethod
    def _get_steps_from(item):
        if hasattr(item, 'test_steps'):
            return item.test_steps
        return []

    @staticmethod
    def _get_step_results_from(item):
        if hasattr(item, 'test_results_steps'):
            return item.test_results_steps
        return []

    @staticmethod
    def _get_setup_from(item):
        if hasattr(item, 'test_method_setup_steps'):
            method_setup_steps = item.test_method_setup_steps
            del item.test_method_setup_steps
            return method_setup_steps
        return []

    @staticmethod
    def _get_setup_results_from(item):
        if hasattr(item, 'test_method_setup_results_steps'):
            method_setup_results_steps = item.test_method_setup_results_steps
            del item.test_method_setup_results_steps
            return method_setup_results_steps
        return []

    @staticmethod
    def _get_teardown_from(item):
        if hasattr(item, 'test_method_teardown_steps'):
            method_teardown_steps = item.test_method_teardown_steps
            del item.test_method_teardown_steps
            return method_teardown_steps
        return []

    @staticmethod
    def _get_teardown_results_from(item):
        if hasattr(item, 'test_method_teardown_results_steps'):
            method_teardown_results_steps = item.test_method_teardown_results_steps
            del item.test_method_teardown_results_steps
            return method_teardown_results_steps
        return []

    @staticmethod
    def _get_result_links_from(item):
        if hasattr(item, 'result_links'):
            result_links = item.result_links
            del item.result_links
            return result_links
        return []

    @staticmethod
    def _get_duration_from(item):
        if hasattr(item, 'test_duration'):
            test_duration = item.test_duration
            del item.test_duration
            return test_duration
        return None

    @staticmethod
    def _get_traces_from(item):
        if hasattr(item, 'test_traces'):
            test_traces = item.test_traces
            del item.test_traces
            return test_traces
        return None

    @staticmethod
    def _get_attachments_from(item):
        if hasattr(item, 'test_attachments'):
            test_attachments = item.test_attachments
            del item.test_attachments
            return test_attachments
        return []

    @staticmethod
    def _get_parameters_from(item):
        if hasattr(item, 'test_parameters'):
            return item.test_parameters
        return None

    @staticmethod
    def _get_classname_from(item):
        i = item.function.__qualname__.find('.')
        if i != -1:
            return item.function.__qualname__[:i]
        return None

    @staticmethod
    def _set_links(item, data):
        if hasattr(item, 'array_parametrize_mark_id'):
            for link in item.function.test_links:
                data['links'].append({})
                data['links'][-1]['url'] = TestITListener.attribute_collector_links(
                                                link,
                                                'url',
                                                item.own_markers,
                                                item.array_parametrize_mark_id,
                                                item.index)
                if link['title']:
                    data['links'][-1]['title'] = TestITListener.attribute_collector_links(
                                                    link,
                                                    'title',
                                                    item.own_markers,
                                                    item.array_parametrize_mark_id,
                                                    item.index)
                if link['type']:
                    data['links'][-1]['type'] = TestITListener.attribute_collector_links(
                                                    link,
                                                    'type',
                                                    item.own_markers,
                                                    item.array_parametrize_mark_id,
                                                    item.index)
                if link['description']:
                    data['links'][-1][
                        'description'] = TestITListener.attribute_collector_links(
                                            link,
                                            'description',
                                            item.own_markers,
                                            item.array_parametrize_mark_id,
                                            item.index)
        else:
            data['links'] = item.function.test_links

    @staticmethod
    def _get_title_from(item):
        if not hasattr(item.function, 'test_title'):
            return None
        if hasattr(item, 'array_parametrize_mark_id'):
            return TestITListener.param_attribute_collector(item.function.test_title,
                                                            item.own_markers,
                                                            item.array_parametrize_mark_id,
                                                            item.index)
        return item.function.test_title

    @staticmethod
    def _get_description_from(item):
        if not hasattr(item.function, 'test_description'):
            return None
        if hasattr(item, 'array_parametrize_mark_id'):
            return TestITListener.param_attribute_collector(
                item.function.test_description,
                item.own_markers,
                item.array_parametrize_mark_id,
                item.index)
        return item.function.test_description

    @staticmethod
    def _set_labels(item, data):
        if hasattr(item, 'array_parametrize_mark_id'):
            result, param_id = TestITListener.mass_param_attribute_collector(
                item.function.test_labels[0],
                item.own_markers,
                item.array_parametrize_mark_id,
                item.index)
            if param_id is not None and item.function.test_labels[0][1:-1] in \
                    item.name[(item.name.find('[') + 1):(item.name.rfind(']'))].split(
                        '-')[param_id]:
                for label in result:
                    data['labels'].append({
                        'name': label
                    })
            else:
                data['labels'].append({
                    'name': result
                })
        else:
            for label in item.function.test_labels:
                data['labels'].append({
                    'name': label
                })

    @staticmethod
    def _set_workitems_id(item, data):
        if hasattr(item, 'array_parametrize_mark_id'):
            result, param_id = TestITListener.mass_param_attribute_collector(
                item.function.test_workitems_id[0], item.own_markers,
                item.array_parametrize_mark_id, item.index)
            if param_id is not None and item.function.test_workitems_id[0][
                                        1:-1] in \
                    item.name[(item.name.find('[') + 1):(item.name.rfind(']'))].split(
                        '-')[param_id]:
                data['workItemsID'] = result
            else:
                data['workItemsID'] = [result]
        else:
            data['workItemsID'] = item.function.test_workitems_id

    @staticmethod
    def attribute_collector(item):
        # TODO: need to refactoring that method
        if not hasattr(item.function, 'test_external_id'):
            return None
        data = {
            'externalID': item.test_external_id,
            'configurationID': item.test_configuration_id,
            'steps': TestITListener._get_steps_from(item),
            'stepResults': TestITListener._get_step_results_from(item),
            'setUp': TestITListener._get_setup_from(item),
            'setUpResults': TestITListener._get_setup_results_from(item),
            'tearDown': TestITListener._get_teardown_from(item),
            'tearDownResults': TestITListener._get_teardown_results_from(
                item),
            'resultLinks': TestITListener._get_result_links_from(item),
            'duration': TestITListener._get_duration_from(item),
            'testResult': 'Failed' if hasattr(item,
                                              'test_traces') else 'Passed',
            'failureReasonName': 'TestDefect' if hasattr(item,
                                                         'test_traces') else 'NoDefect',
            'traces': TestITListener._get_traces_from(item),
            'namespace': item.function.__module__,
            'attachments': TestITListener._get_attachments_from(item),
            'parameters': TestITListener._get_parameters_from(item),
            'classname': TestITListener._get_classname_from(item),
            'title': TestITListener._get_title_from(item),
            'description': TestITListener._get_description_from(item),
            'links': [],
            'labels': [],
            'workItemsID': [],
            'message': None
        }

        if hasattr(item.function, 'test_links'):
            TestITListener._set_links(item, data)

        if hasattr(item.function, 'test_labels'):
            TestITListener._set_labels(item, data)

        if hasattr(item.function, 'test_workitems_id'):
            TestITListener._set_workitems_id(item, data)

        if item.own_markers:
            for mark in item.own_markers:
                if mark.name == 'skip' or mark.name == 'skipif' and mark.args[0]:
                    data['testResult'] = 'Skipped'
                    data['failureReasonName'] = None
                    if mark.kwargs:
                        data['message'] = mark.kwargs['reason']

        if hasattr(item, 'test_message'):
            data['message'] = item.test_message

        if hasattr(item.function, 'test_displayname'):
            if hasattr(item, 'array_parametrize_mark_id'):
                data['autoTestName'] = TestITListener.param_attribute_collector(
                    item.function.test_displayname,
                    item.own_markers,
                    item.array_parametrize_mark_id,
                    item.index)
            else:
                data['autoTestName'] = item.function.test_displayname
        elif item.function.__doc__:
            data['autoTestName'] = item.function.__doc__
        else:
            data['autoTestName'] = 'Test without name'
            data['testResult'] = 'Failed'
            data['failureReasonName'] = 'TestDefect'
            data['traces'] = '>\n{}\nE {} must have @testit.displayName or documentation!\n{}:{}: Exception'.format(
                inspect.getsource(item.function),
                item.originalname,
                item.location[0],
                item.location[1])

        return data

    @staticmethod
    def param_attribute_collector(attribute, marks, parametrize_id, index):
        for ID in parametrize_id:
            if attribute[attribute.find('{') + 1:attribute.rfind('}')] in \
                    marks[ID].args[0]:
                return attribute.split('{')[0] + marks[ID].args[1][index][
                    marks[ID].args[0].split(', ').index(attribute[
                                                        attribute.find(
                                                            '{') + 1:attribute.rfind(
                                                            '}')])] + \
                       attribute.split('}')[1]
        return attribute

    @staticmethod
    def mass_param_attribute_collector(attribute, marks, parametrize_id, index):
        for ID in parametrize_id:
            if attribute[1:-1] != '' and attribute[1:-1] in marks[ID].args[0]:
                param_id = marks[ID].args[0].split(', ').index(attribute[1:-1])
                return marks[ID].args[1][index][param_id], param_id
        return attribute, None

    @staticmethod
    def attribute_collector_links(link, key, marks, parametrize_id, index):
        for ID in parametrize_id:
            if link[key][(link[key].find('{') + 1):(link[key].rfind('}'))] in \
                    marks[ID].args[0]:
                return link[key].split('{')[0] + marks[ID].args[1][index][
                    marks[ID].args[0].split(', ').index(link[key][(
                                                                          link[key].find('{') + 1):
                                                                  (link[key].rfind(
                                                                      '}'))])] + \
                       link[key].split('}')[1]
        return link[key]

    @staticmethod
    def form_tree_steps(item, tree_steps, stage):
        for scope in ('session', 'module', 'class', 'method'):
            if hasattr(item, f'test_{scope}_{stage}_steps'):
                tree_steps[scope] = getattr(
                    item, f'test_{scope}_{stage}_steps')
                tree_steps[f'{scope}_results'] = getattr(
                    item, f'test_{scope}_{stage}_results_steps')
            elif scope not in tree_steps:
                tree_steps[scope] = []
                tree_steps[f'{scope}_results'] = []
        return tree_steps

    @testit_pytest.hookimpl
    def add_link(self, link_url, link_title, link_type, link_description):
        if not hasattr(self.item, 'result_links'):
            self.item.result_links = []
        self.item.result_links.append(
            {
                'url': link_url,
                'title': link_title,
                'type': link_type,
                'description': link_description
            })

    @testit_pytest.hookimpl
    def add_message(self, test_message):
        self.item.test_message = str(test_message)

    @testit_pytest.hookimpl
    def add_attachments(self, attach_paths):
        if not hasattr(self.item, 'test_attachments'):
            self.item.test_attachments = []
        self.item.test_attachments += self.load_attachments(attach_paths)

    @testit_pytest.hookimpl
    def load_attachments(self, attach_paths):
        attachments = []
        for path in attach_paths:
            if os.path.isfile(path):
                attachments.append(
                    {
                        'id': self.requests.load_attachment(open(path, "rb"))
                    })
            else:
                print(f'File ({path}) not found!')
        return attachments

    @testit_pytest.hookimpl
    def add_parameters(self, parameters):
        if not hasattr(self.item, 'test_parameters'):
            self.item.test_parameters = {}
            for key, parameter in parameters.items():
                self.item.test_parameters[key] = str(parameter)

    @testit_pytest.hookimpl
    def get_pytest_check_outcome(self):
        if hasattr(self, 'pytest_check_get_failures'):
            failures = self.pytest_check_get_failures()
            if failures and ((not hasattr(self.item, 'old_failures')) or self.item.old_failures != failures):
                self.item.old_failures = failures[:]
                return 'Failed'
        return 'Passed'



import inspect
import os
import re
from datetime import datetime
from functools import wraps

from testit_pytest import TestITPluginManager


def inner(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        if hasattr(TestITPluginManager.get_plugin_manager().hook, 'add_parameters') and kwargs:
            TestITPluginManager.get_plugin_manager().hook.add_parameters(parameters=kwargs)
        function(*args, **kwargs)
        return function
    return wrapper


def workItemID(*test_workitems_id: int or str):
    def outer(function):
        function.test_workitems_id = []
        for test_workitem_id in test_workitems_id:
            function.test_workitems_id.append(str(test_workitem_id))
        return inner(function)
    return outer


def displayName(test_displayname: str):
    def outer(function):
        function.test_displayname = test_displayname
        return inner(function)
    return outer


def externalID(test_external_id: str):
    def outer(function):
        function.test_external_id = test_external_id
        return inner(function)
    return outer


def title(test_title: str):
    def outer(function):
        function.test_title = test_title
        return inner(function)
    return outer


def description(test_description: str):
    def outer(function):
        function.test_description = test_description
        return inner(function)
    return outer


def labels(*test_labels: str):
    def outer(function):
        function.test_labels = test_labels
        return inner(function)
    return outer


def link(url: str, title: str = None, type: str = None, description: str = None):
    def outer(function):
        if not hasattr(function, 'test_links'):
            function.test_links = []
        function.test_links.append({'url': url, 'title': title, 'type': type, 'description': description})
        return inner(function)
    return outer


class LinkType:
    RELATED = 'Related'
    BLOCKED_BY = 'BlockedBy'
    DEFECT = 'Defect'
    ISSUE = 'Issue'
    REQUIREMENT = 'Requirement'
    REPOSITORY = 'Repository'


def addLink(url: str, title: str = None, type: str = None, description: str = None):
    if hasattr(TestITPluginManager.get_plugin_manager().hook, 'add_link'):
        TestITPluginManager.get_plugin_manager().hook.add_link(link_url=url, link_title=title, link_type=type, link_description=description)


def message(test_message: str):
    if hasattr(TestITPluginManager.get_plugin_manager().hook, 'add_message'):
        TestITPluginManager.get_plugin_manager().hook.add_message(test_message=test_message)


def attachments(*attachments_paths: str):
    if step.step_is_active():
        step.add_attachments(attachments_paths)
    else:
        if hasattr(TestITPluginManager.get_plugin_manager().hook, 'add_attachments'):
            TestITPluginManager.get_plugin_manager().hook.add_attachments(attach_paths=attachments_paths)


class step:
    step_stack = []
    steps_data = []
    steps_data_results = []
    attachments = []

    def __init__(self, *args, **kwargs):
        self.args = args
        if 'parameters' in kwargs and kwargs['parameters']:
            self.parameters = kwargs['parameters']

    def __call__(self, *args, **kwargs):
        parameters = {}
        if self.args and callable(self.args[0]):
            function = self.args[0]

            if self.step_stack:
                name = f'Step {str(self.step_stack[0] + 1)}'
                steps = self.steps_data[self.step_stack[0]]['steps']
                for step_id in self.step_stack[1:]:
                    name += f'.{step_id + 1}'
                    steps = steps[step_id]['steps']
                name += f'.{len(steps) + 1}'
            else:
                name = f'Step {str(len(self.steps_data) + 1)}'

            args_default_values = inspect.getfullargspec(function).defaults
            if args or args_default_values:
                if args_default_values:
                    args += args_default_values

                function_args_name = inspect.getfullargspec(function).args
                step_args = [arg_name for arg_name in function_args_name if arg_name not in list(kwargs)]

                for id in range(0, len(step_args)):
                    parameters[step_args[id]] = str(args[id])
            if kwargs:
                for key, parameter in kwargs.items():
                    parameters[key] = str(parameter)

            with step(name, function.__name__, parameters=parameters):
                return function(*args, **kwargs)
        else:
            function = args[0]

            @wraps(function)
            def step_wrapper(*a, **kw):
                if self.args:
                    args_default_values = inspect.getfullargspec(function).defaults
                    if a or args_default_values:
                        if args_default_values:
                            a += args_default_values

                        function_args_name = inspect.getfullargspec(function).args
                        step_args = [arg_name for arg_name in function_args_name if arg_name not in list(kw)]

                        for id in range(0, len(step_args)):
                            parameters[step_args[id]] = str(a[id])
                    if kw:
                        for key, parameter in kw.items():
                            parameters[key] = str(parameter)

                    with step(
                            self.args[0], self.args[1], parameters=parameters) if len(self.args) == 2\
                            else step(self.args[0], parameters=parameters
                        ):
                        return function(*a, **kw)
            return step_wrapper

    def __enter__(self):
        self.start_time = round(datetime.utcnow().timestamp() * 1000)
        self.steps_data = self.step_append(
                            self.steps_data,
                            self.step_stack,
                            self.args[0],
                            self.args[1] if len(self.args) == 2 else None)

    def __exit__(self, exc_type, exc_value, tb):
        outcome = 'Failed' if exc_type else TestITPluginManager.get_plugin_manager().hook.get_pytest_check_outcome()[0] if\
            hasattr(TestITPluginManager.get_plugin_manager().hook, 'get_pytest_check_outcome') else 'Passed'
        duration = round(datetime.utcnow().timestamp() * 1000) - self.start_time
        self.steps_data_results = self.result_step_append(
                                    self.steps_data,
                                    self.steps_data_results,
                                    self.step_stack,
                                    outcome,
                                    duration)

    def step_append(self, steps, step_stack, step_title, step_description):
        if step_stack:
            steps[step_stack[0]]['steps'] = self.step_append(steps[step_stack[0]]['steps'], step_stack[1:], step_title, step_description)
        else:
            steps.append({'title': step_title, 'description': step_description, 'steps': []})
            self.step_stack.append(len(steps) - 1)
            self.attachments.append([])
        return steps

    def result_step_append(self, steps, steps_results, step_stack, outcome, duration):
        if len(step_stack) == 1:
            while len(steps_results) < step_stack[0] + 1:
                steps_results.append({})
            steps_results[step_stack[0]]['title'] = steps[step_stack[0]]['title']
            steps_results[step_stack[0]]['description'] = steps[step_stack[0]]['description']
            steps_results[step_stack[0]]['outcome'] = outcome
            steps_results[step_stack[0]]['duration'] = duration
            steps_results[step_stack[0]]['parameters'] = self.parameters if hasattr(self, 'parameters') else None
            steps_results[step_stack[0]]['attachments'] = self.attachments[-1]
            del self.step_stack[-1]
            del self.attachments[-1]
        else:
            while len(steps_results) < step_stack[0] + 1:
                steps_results.append({'stepResults': []})
            steps_results[step_stack[0]]['stepResults'] = self.result_step_append(
                                                                steps[step_stack[0]]['steps'],
                                                                steps_results[step_stack[0]]['stepResults'],
                                                                step_stack[1:],
                                                                outcome,
                                                                duration)
        return steps_results

    @classmethod
    def get_steps_data(cls):
        data = cls.steps_data
        result_data = cls.steps_data_results
        cls.steps_data = []
        cls.steps_data_results = []
        return data, result_data

    @classmethod
    def step_is_active(cls):
        return len(cls.step_stack) != 0

    @classmethod
    def add_attachments(cls, attachments_paths):
        if hasattr(TestITPluginManager.get_plugin_manager().hook, 'load_attachments'):
            cls.attachments[-1] += TestITPluginManager.get_plugin_manager().hook.load_attachments(attach_paths=attachments_paths)[0]


def search_in_environ(variable):
    if re.fullmatch(r'{[a-zA-Z_]\w*}', variable) and variable[1:-1] in os.environ:
        return os.environ[variable[1:-1]]
    return variable


def configurations_parser(data_autotests):
    return {
        data_autotest['autoTest']['externalId']: data_autotest['configurationId']
        for data_autotest in data_autotests
    }


def uuid_check(uuid):
    if not re.fullmatch(r'[a-zA-Z0-9]{8}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{12}', uuid):
        print(f'The wrong {uuid}!')
        raise SystemExit
    return uuid


def url_check(url):
    if not re.fullmatch(r'^(?:(?:(?:https?|ftp):)?//)?(?:(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))|(?:(?:[a-zA-Z0-9\u00a1-\uffff][a-zA-Z0-9\u00a1-\uffff_-]{0,62})?[a-zA-Z0-9\u00a1-\uffff]\.)+(?:[a-zA-Z\u00a1-\uffff]{2,}\.?))(?::\d{2,5})?(?:[/?#]\S*)?$', url):
        print('The wrong URL!')
        raise SystemExit
    if url[-1] == '/':
        return url[:-1]
    return url



from testit_pytest.utils import (
    LinkType,
    addLink,
    attachments,
    description,
    displayName,
    externalID,
    labels,
    link,
    message,
    step,
    title,
    workItemID
)

__all__ = [
    'workItemID',
    'displayName',
    'externalID',
    'title',
    'description',
    'labels',
    'link',
    'addLink',
    'message',
    'attachments',
    'LinkType',
    'step'
]



class JSONFixture:

    @staticmethod
    def create_autotest(
        external_id,
        project_id,
        name,
        namespace,
        classname,
        links,
        steps,
        setup,
        teardown,
        title,
        description,
        labels
    ):
        return {
            'externalId': external_id,
            'projectId': project_id,
            'name': name,
            'namespace': namespace,
            'classname': classname,
            'links': links,
            'steps': steps,
            'setup': setup,
            'teardown': teardown,
            'title': title,
            'description': description,
            'labels': labels
        }

    @staticmethod
    def update_autotest(
        external_id,
        project_id,
        name,
        namespace,
        classname,
        links,
        steps,
        setup,
        teardown,
        title,
        description,
        labels,
        ID
    ):
        return {
            'externalId': external_id,
            'projectId': project_id,
            'name': name,
            'namespace': namespace,
            'classname': classname,
            'links': links,
            'steps': steps,
            'setup': setup,
            'teardown': teardown,
            'title': title,
            'description': description,
            'labels': labels,
            'id': ID
        }

    @staticmethod
    def create_testrun(project_id, name):
        return {
            'projectId': project_id,
            'name': name
        }

    @staticmethod
    def set_results_for_testrun(
        autotest_external_id,
        configuration_id,
        outcome,
        step_results,
        setup_results,
        teardown_results,
        traces,
        links,
        duration,
        failure_reason_name,
        message,
        parameters,
        attachments
    ):
        return {
            'configurationId': configuration_id,
            'links': links,
            'autoTestExternalId': autotest_external_id,
            'outcome': outcome,
            'traces': traces,
            'stepResults': step_results,
            'setupResults': setup_results,
            'teardownResults': teardown_results,
            'duration': duration,
            'failureReasonName': failure_reason_name,
            'message': message,
            'parameters': parameters,
            'attachments': attachments
        }



from setuptools import setup

setup(
    name='testit-pytest',
    version='0.2.10',
    description='Pytest plugin for Test IT',
    long_description=open('README.rst').read(),
    url='https://pypi.org/project/testit-pytest/',
    author='Pavel Butuzov',
    author_email='pavel.butuzov@testit.software',
    license='Apache-2.0',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    py_modules=['testit', 'testit_pytest'],
    packages=['testit_pytest'],
    package_dir={'testit_pytest': 'src'},
    install_requires=['pytest', 'requests', 'pytest-xdist'],
    entry_points={'pytest11': ['testit_pytest = testit_pytest.plugin']}
)



import testit_pytest
import pytest
from testit_pytest.listener import TestITListener


def pytest_addoption(parser):
    parser.getgroup('testit').addoption(
        '--testit',
        action='store_true',
        dest="testit_report",
        help='Pytest plugin for Test IT'
    )
    parser.getgroup('testit').addoption(
        '--testrunid',
        action="store",
        dest="set_testrun",
        metavar="5236eb3f-7c05-46f9-a609-dc0278896464",
        help='Set Test-run ID'
    )
    parser.getgroup('testit').addoption(
        '--testit_url',
        action="store",
        dest="set_url",
        metavar="https://demo.testit.software",
        help='Set location of the Test IT instance'
    )
    parser.getgroup('testit').addoption(
        '--privatetoken',
        action="store",
        dest="set_privatetoken",
        metavar="T2lKd2pLZGI4WHRhaVZUejNl",
        help='Set API secret key'
    )
    parser.getgroup('debug').addoption(
        '--testit_proxy',
        action="store",
        dest="set_testit_proxy",
        metavar='{"http":"http://localhost:8888","https":"http://localhost:8888"}',
        help='Set proxy for sending requests'
    )


@pytest.mark.tryfirst
def pytest_cmdline_main(config):
    if config.option.testit_report:
        listener = TestITListener(config.option.set_testrun,
                                  config.option.set_url,
                                  config.option.set_privatetoken,
                                  config.option.set_testit_proxy)
        config.pluginmanager.register(listener)
        testit_pytest.TestITPluginManager.get_plugin_manager().register(listener)
